const core = @import("core");
const std = @import("std");

///
/// An entry block of an Oar archive file.
///
/// Typically, following the block in memory is the file data it holds the meta-information for.
///
pub const Entry = extern struct {
    signature: [signature_magic.len]u8 = signature_magic,
    revision: u8,
    path: Path,
    file_size: u64,
    absolute_offset: u64,
    padding: [232]u8 = std.mem.zeroes([232]u8),

    comptime {
        const entry_size = @sizeOf(Entry);

        if (entry_size != 512)
            @compileError("Entry is " ++
                std.fmt.comptimePrint("{d}", .{entry_size}) ++ " bytes");
    }

    ///
    /// Attempts to read the next [Entry] from `file_access`.
    ///
    /// Returns the read [Entry], `null` if there is no more to read, or a
    /// [core.io.FileAccess.Error] if it failed.
    ///
    pub fn next(file_access: core.io.FileAccess) core.io.FileAccess.Error!?Entry {
        const mem = std.mem;
        var entry = mem.zeroes(Entry);
        const origin = try file_access.queryCursor();

        if (((try file_access.read(mem.asBytes(&entry))) != @sizeOf(Entry)) and
            core.io.equalsBytes(&entry.signature, &signature_magic)) {

            try file_access.seek(origin);

            return null;
        }

        return entry;
    }

    ///
    /// Magic identifier used to validate [Entry] data.
    ///
    pub const signature_magic = [3]u8{'o', 'a', 'r'};
};

///
/// Unique identifier pointing to an entry within an archive.
///
/// A path does not do any verification that the given entry pointed to actually exists.
///
pub const Path = extern struct {
    buffer: [255]u8,
    length: u8,

    ///
    /// [Error.TooLong] occurs when creating a path that is greater than the maximum path size **in
    /// bytes**.
    ///
    pub const Error = error {
        TooLong,
    };

    ///
    /// An empty [Path] with a length of `0`.
    ///
    pub const empty = std.mem.zeroes(Path);

    ///
    /// Returns `true` if `this_path` is equal to `that_path, otherwise `false`.
    ///
    pub fn equals(this_path: Path, that_path: Path) bool {
        return core.io.equalsBytes(this_path.buffer[0 ..this_path.
            length], that_path.buffer[0 .. that_path.length]);
    }

    ///
    /// Returns the hash of the text in `path`.
    ///
    pub fn hash(path: Path) usize {
        return core.io.hashBytes(path.buffer[0 .. path.length]);
    }

    ///
    /// Attempts to create a [Path] with the path components in `sequences` as a fully qualified
    /// path from root.
    ///
    /// A [Path] value is returned containing the fully qualified path from the file-system root or
    /// a [Error] if it could not be created.
    ///
    pub fn joined(sequences: []const []const u8) Error!Path {
        var path = empty;

        if (sequences.len != 0) {
            const last_sequence_index = sequences.len - 1;

            for (sequences) |sequence, index| if (sequence.len != 0) {
                var components = core.io.Spliterator(u8){
                    .source = sequence,
                    .delimiter = "/",
                };

                while (components.next()) |component| if (component.len != 0) {
                    for (component) |byte| {
                        if (path.length == max) return error.TooLong;

                        path.buffer[path.length] = byte;
                        path.length += 1;
                    }

                    if (components.hasNext()) {
                        if (path.length == max) return error.TooLong;

                        path.buffer[path.length] = '/';
                        path.length += 1;
                    }
                };

                if (index < last_sequence_index) {
                    if (path.length == max) return error.TooLong;

                    path.buffer[path.length] = '/';
                    path.length += 1;
                }
            };
        }

        return path;
    }

    ///
    /// Maximum number of **bytes** in a [Path].
    ///
    pub const max = 255;

    ///
    /// Textual separator between components of a [Path].
    ///
    pub const seperator = '/';
};

///
/// Starts the **O**na **Ar**chive packer utility.
///
pub fn main() !u8 {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};

    defer _ = gpa.deinit();

    const process = std.process;
    const allocator = gpa.allocator();
    const args = try process.argsAlloc(allocator);

    defer process.argsFree(allocator, args);

    const outWriter = std.io.getStdOut().writer();

    if (args.len > 1) {
        const command = args[1];
        const io = core.io;

        if (io.equalsBytes(command, "pack")) {
            return 0;
        }

        if (io.equalsBytes(command, "unpack")) {
            return 0;
        }

        try outWriter.print("Unknown command: {s}", .{command});

        return 1;
    }

    try outWriter.print("{s}", .{args[0]});

    return 0;
}

test {

}
